# -*- coding: utf-8 -*-
"""
    DISCLAIMER: I BLATANTLY STOLE THIS FROM FLASK BECAUSE I LIKED IT SO MUCH:

    Copyright (c) 2015 by Armin Ronacher and contributors.  See AUTHORS
    for more details.

    Some rights reserved.

    Redistribution and use in source and binary forms of the software as well
    as documentation, with or without modification, are permitted provided
    that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    * The names of the contributors may not be used to endorse or
      promote products derived from this software without specific
      prior written permission.

    THIS SOFTWARE AND DOCUMENTATION IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
    NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE AND DOCUMENTATION, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
    DAMAGE.

    flask.config
    ~~~~~~~~~~~~
    Implements the configuration related objects.
    :copyright: (c) 2015 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""

import os
import sys
import pkgutil
import types
import errno

from werkzeug.utils import import_string
import json

PY2 = sys.version_info[0] == 2
_identity = lambda x: x

if not PY2:
    text_type = str
    string_types = (str,)
    integer_types = (int,)

    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

    from io import StringIO

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    implements_to_string = _identity

else:
    text_type = unicode
    string_types = (str, unicode)
    integer_types = (int, long)

    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()

    from cStringIO import StringIO

    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls

def get_root_path(import_name):
    """Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.
    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, '__file__', None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError('No root path can be found for the provided '
                               'module "%s".  This can happen because the '
                               'module came from an import hook that does '
                               'not provide file name information or because '
                               'it\'s a namespace package.  In this case '
                               'the root path needs to be explicitly '
                               'provided.' % import_name)

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))

class ConfigAttribute(object):
    """Makes an attribute forward to the config"""

    def __init__(self, name, get_converter=None):
        self.__name__ = name
        self.get_converter = get_converter

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        rv = obj.config[self.__name__]
        if self.get_converter is not None:
            rv = self.get_converter(rv)
        return rv

    def __set__(self, obj, value):
        obj.config[self.__name__] = value


class Config(dict):
    """Works exactly like a dict but provides ways to fill it from files
    or special dictionaries.  There are two common patterns to populate the
    config.
    Either you can fill the config from a config file::
        app.config.from_pyfile('yourconfig.cfg')
    Or alternatively you can define the configuration options in the
    module that calls :meth:`from_object` or provide an import path to
    a module that should be loaded.  It is also possible to tell it to
    use the same module and with that provide the configuration values
    just before the call::
        DEBUG = True
        SECRET_KEY = 'development key'
        app.config.from_object(__name__)
    In both cases (loading from any Python file or loading from modules),
    only uppercase keys are added to the config.  This makes it possible to use
    lowercase values in the config file for temporary values that are not added
    to the config or to define the config keys in the same file that implements
    the application.
    Probably the most interesting way to load configurations is from an
    environment variable pointing to a file::
        app.config.from_envvar('YOURAPPLICATION_SETTINGS')
    In this case before launching the application you have to set this
    environment variable to the file you want to use.  On Linux and OS X
    use the export statement::
        export YOURAPPLICATION_SETTINGS='/path/to/config/file'
    On windows use `set` instead.
    :param root_path: path to which files are read relative from.  When the
                      config object is created by the application, this is
                      the application's :attr:`~flask.Flask.root_path`.
    :param defaults: an optional dictionary of default values
    """

    def __init__(self, root_path, defaults=None):
        dict.__init__(self, defaults or {})
        self.root_path = root_path

    def from_envvar(self, variable_name, silent=False):
        """Loads a configuration from an environment variable pointing to
        a configuration file.  This is basically just a shortcut with nicer
        error messages for this line of code::
            app.config.from_pyfile(os.environ['YOURAPPLICATION_SETTINGS'])
        :param variable_name: name of the environment variable
        :param silent: set to ``True`` if you want silent failure for missing
                       files.
        :return: bool. ``True`` if able to load config, ``False`` otherwise.
        """
        rv = os.environ.get(variable_name)
        if not rv:
            if silent:
                return False
            raise RuntimeError('The environment variable %r is not set '
                               'and as such configuration could not be '
                               'loaded.  Set this variable and make it '
                               'point to a configuration file' %
                               variable_name)
        return self.from_pyfile(rv, silent=silent)

    def from_pyfile(self, filename, silent=False):
        """Updates the values in the config from a Python file.  This function
        behaves as if the file was imported as module with the
        :meth:`from_object` function.
        :param filename: the filename of the config.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to ``True`` if you want silent failure for missing
                       files.
        .. versionadded:: 0.7
           `silent` parameter.
        """
        filename = os.path.join(self.root_path, filename)
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        """Updates the values from the given object.  An object can be of one
        of the following two types:
        -   a string: in this case the object with that name will be imported
        -   an actual object reference: that object is used directly
        Objects are usually either modules or classes.
        Just the uppercase variables in that object are stored in the config.
        Example usage::
            app.config.from_object('yourapplication.default_config')
            from yourapplication import default_config
            app.config.from_object(default_config)
        You should not use this function to load the actual configuration but
        rather configuration defaults.  The actual config should be loaded
        with :meth:`from_pyfile` and ideally from a location not within the
        package because the package might be installed system wide.
        :param obj: an import name or object
        """
        if isinstance(obj, string_types):
            obj = import_string(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def from_json(self, filename, silent=False):
        """Updates the values in the config from a JSON file. This function
        behaves as if the JSON object was a dictionary and passed to the
        :meth:`from_mapping` function.
        :param filename: the filename of the JSON file.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to ``True`` if you want silent failure for missing
                       files.
        .. versionadded:: 1.0
        """
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as json_file:
                obj = json.loads(json_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj)

    def from_mapping(self, *mapping, **kwargs):
        """Updates the config like :meth:`update` ignoring items with non-upper
        keys.
        .. versionadded:: 1.0
        """
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(mapping[0].items())
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        mappings.append(kwargs.items())
        for mapping in mappings:
            for (key, value) in mapping:
                if key.isupper():
                    self[key] = value
        return True

    def get_namespace(self, namespace, lowercase=True, trim_namespace=True):
        """Returns a dictionary containing a subset of configuration options
        that match the specified namespace/prefix. Example usage::
            app.config['IMAGE_STORE_TYPE'] = 'fs'
            app.config['IMAGE_STORE_PATH'] = '/var/app/images'
            app.config['IMAGE_STORE_BASE_URL'] = 'http://img.website.com'
            image_store_config = app.config.get_namespace('IMAGE_STORE_')
        The resulting dictionary `image_store` would look like::
            {
                'type': 'fs',
                'path': '/var/app/images',
                'base_url': 'http://img.website.com'
            }
        This is often useful when configuration options map directly to
        keyword arguments in functions or class constructors.
        :param namespace: a configuration namespace
        :param lowercase: a flag indicating if the keys of the resulting
                          dictionary should be lowercase
        :param trim_namespace: a flag indicating if the keys of the resulting
                          dictionary should not include the namespace
        .. versionadded:: 1.0
        """
        rv = {}
        for k, v in iteritems(self):
            if not k.startswith(namespace):
                continue
            if trim_namespace:
                key = k[len(namespace):]
            else:
                key = k
            if lowercase:
                key = key.lower()
            rv[key] = v
        return rv

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, dict.__repr__(self))



from calamari.interfaces import CalamariInterface
from twisted.words.protocols import irc
from twisted.internet import reactor, protocol, ssl
from threading import Thread

class IRCBot( irc.IRCClient ):
    def connectionMade(self):
        irc.IRCClient.connectionMade(self)

    def connectionLost(self, reason):
        irc.IRCClient.connectionLost(self, reason)

    def signedOn(self):
        """Called when bot has succesfully signed on to server."""
        self.join(self.factory.channel)

    def privmsg(self, user, channel, msg):
        """This will get called when the bot receives a message."""
        user = user.split('!', 1)[0]
        
        # Check to see if they're sending me a private message
        if channel == self.nickname:
            msg = "It isn't nice to whisper!  Play nice with the group."
            self.msg(user, msg)
            return

        self.interface.process( {'message':msg, 'user':user} )

    # For fun, override the method that determines how a nickname is changed on
    # collisions. The default method appends an underscore.
    def alterCollidedNick(self, nickname):
        """
        Generate an altered version of a nickname that caused a collision in an
        effort to create an unused related name for subsequent registration.
        """
        return nickname + '^'

    def threadSafeMsg(self, channel, msg ):
        reactor.callFromThread( self.msg, channel, msg )

class IRCBotFactory(protocol.ClientFactory):
    """A factory for IRCBots

    A new protocol instance will be created each time we connect to the server.
    """

    def __init__(self, interface):
        self.interface = interface
        self.channel = interface.channel

    def buildProtocol(self, addr):
        p = IRCBot()
        p.nickname = self.interface.nickname
        p.username = self.interface.username
        p.password = self.interface.password
        p.interface = self.interface
        self.interface.bot_instance = p
        p.factory = self
        return p

    def clientConnectionLost(self, connector, reason):
        """If we get disconnected, reconnect to server."""
        connector.connect()

    def clientConnectionFailed(self, connector, reason):
        reactor.stop()

class CalamariIRC( CalamariInterface ):
    def __init__( self, irc_server, irc_port, channel, nickname,
            use_ssl = False, username = None, password = None ):
        """
        Initialize the irc interface
        """
        self.irc_server = irc_server
        self.irc_port = irc_port
        self.channel = channel
        self.nickname = nickname
        self.use_ssl = use_ssl
        self.username = username
        self.password = password

    def write_output( self, msg ):
        """
        writes message to the irc channel
        """
        self.bot_instance.threadSafeMsg("#" + self.channel, str(msg['message']))

    def run( self ):
        """
        start the main server thread
        """
        t = Thread( target = self.server_thread )
        t.daemon = True
        t.start()

    def server_thread( self ):
        f = IRCBotFactory( self )
        if self.use_ssl:
            reactor.connectSSL( self.irc_server, self.irc_port, f,
                    ssl.ClientContextFactory() )
        else:
            reactor.connectTCP( self.irc_server, self.irc_port, f )
        reactor.run(installSignalHandlers=False)



"""
interfaces/xmpp.py

Defines an xmpp interface for the bot
"""
from calamari.interfaces import CalamariInterface
import sys
import sleekxmpp

# The sleekxmpp docs reccomend having this...not very pythonic, might try
# removing it later because YOLO
if sys.version_info < (3, 0):
    from sleekxmpp.util.misc_ops import setdefaultencoding
    setdefaultencoding('utf8')
else:
    raw_input = input

class CalamariXMPP( CalamariInterface ):
    """
    An XMPP interface to the bot
    """
    def __init__(self, jid, password, room, nick):
        """
        Initialize the sleekxmpp object and register all the event handlers and
        plugins we're going to be using
        """
        self.xmpp = sleekxmpp.ClientXMPP(jid, password)
        self.room = room
        self.nick = nick
        self.xmpp.add_event_handler( "session_start", self.start )
        self.xmpp.add_event_handler( "groupchat_message", self.muc_message )
        self.xmpp.register_plugin('xep_0030') # Service Discovery
        self.xmpp.register_plugin('xep_0045') # Multi-User Chat
        self.xmpp.register_plugin('xep_0199') # XMPP Ping

    def start(self, event):
        """
        SleekXMPP event handler.  This gets run once the interface successfully 
        logs in.  Joins the specified chat room.
        """
        self.xmpp.plugin['xep_0045'].joinMUC(self.room, self.nick, wait=True)

    def muc_message(self, msg):
        """
        SleekXMPP event handler.  This gets run every time a new message gets
        put in the chat room.  Note: even the bots messages will trigger this
        handler
        """
        # TODO: Actually define what constitutes a message somewhere
        self.process( {'message':msg['body'], 'user':msg['mucnick']} )

    def write_output( self, msg ):
        """
        Implementation of ABC's abstract write_output method.  Simply writes a
        message to the group chat
        """
        self.xmpp.send_message( mto=self.room, mbody=msg['message'],
                mtype="groupchat" )

    def run( self ):
        """
        Implementation of ABC's abstract run method.  Starts the interface.
        """
        self.xmpp.connect()
        self.xmpp.process( )



"""
interfaces/__init__.py

Defines an abstract base class for the bots interfaces to inherit from
"""
import abc
from Queue import Queue
from threading import Thread

class CalamariInterface(object):
    """
    Abstract interface to the bot.  This class should be extended to implement
    additional interfaces
    """
    __metaclass__ = abc.ABCMeta

    def __new__(cls, *args, **kwargs):
        """
        Fanciness!  Allow some initialization to happen without subclass
        implementors having to explicitly call the parent constructor
        """
        # Instantiate the instance and create the input and output queues
        instance = object.__new__( cls )
        instance._inq = Queue()
        instance._outq = Queue()

        # Create the output thread and start it
        instance.__output_thread = Thread(target=instance.output_handler)
        instance.__output_thread.daemon = True
        instance.__output_thread.start()

        return instance

    def get( self ):
        """
        Get a message from the input queue
        """
        return self._inq.get()

    def put( self, msg ):
        """
        Put a message on the output queue
        """
        self._outq.put( msg )

    def process( self, msg ):
        """
        Put a message on the input queue.  Essentially, process a message
        """
        self._inq.put( msg )

    def output_handler( self ):
        """
        This will be spun up in a separate thread.  Simply consumes messages
        from the output queue and writes them to the interface
        """
        while True:
            self.write_output( self._outq.get() )

    @abc.abstractmethod
    def write_output( self, msg ):
        """
        This method should have the bot simply write a message out to the
        appropriate interface
        """
        pass

    @abc.abstractmethod
    def run( self ):
        """
        This should start the interface.  Ie, it should join the room and begin
        listening
        """
        pass

from calamari.interfaces.xmpp import CalamariXMPP
from calamari.interfaces.irc_interface import CalamariIRC



"""
engine/engine.py

Defines the main logic of the bot, while IO is handled by the interfaces.
"""
from threading import Thread
from config import Config, get_root_path

class CalamariEngine():
    """
    The brains of the operation.  Register interfaces with the engine and it
    will do all of the heavy lifting
    """
    def __init__( self, import_name, prefix = "." ):
        """
        Initialize the list of commands / interfaces, the config dictionary, and
        set the command prefix
        """
        self.import_name = import_name
        self.root_path = get_root_path( import_name )
        self.config = Config( self.root_path )
        self.prefix = prefix
        self.commands = {}
        self.null_cmd = None
        self.interfaces = []

    def register_interface( self, interface, threads = 5 ):
        """
        Registers an interface to the engine and spins up the specified number
        of threads to handle that interface.  
        """
        self.interfaces.append( interface )
        for i in range( threads ):
            t = Thread( target = self.interface_handler, args = (interface,) )
            t.start()

    def interface_handler( self, interface ):
        """
        This is the method run by the interface handler threads.  Gets a
        message from an interface, processes it, and puts the output back into
        the interface, if there was any
        """
        while True:
            output = self.process( interface.get() )
            if output:
                interface.put( {'message':output} )

    def process( self, msg ):
        """
        Determine which command is supposed to be run and do it
        """
        if not msg['message'].startswith( self.prefix ):
            if self.null_cmd:
                return self.null_cmd( msg['message'], msg['user'] )
            return

        command = msg['message'].split()[0][1:]
        if not command:
            return "Usage: .(command) [args]"

        if command in self.commands:
            return self.commands[command]['func']( msg['message'], msg['user'] )

    def null_command( self ):
        """
        A decorator to specify a "default" command for the bot, that gets run
        when the configured prefix is not present
        """
        def wrap( func ):
            self.null_cmd = func
            return func
        return wrap

    def command( self, command_alias, hidden = False ):
        """
        A decorator which maps a prefix command to a method.
        """
        def wrap( func ):
            self.commands[ command_alias ] = {'func':func,'help':func.__doc__,
                    'hidden':hidden}
            return func
        return wrap

    def run( self ):
        """
        Start listening on all the interfaces
        """
        for i in self.interfaces:
            i.run()
